from ezcx.server import Server
from ezcx.server import TagRouter
from ezcx.server import PathRouter

from ezcx.webhooks import WebhookRequest
from ezcx.webhooks import WebhookResponse
from ezcx.webhooks import rich_content


import json


class RichContent:
    
    def __init__(self):
        self.richContent = [[]]
        
    def to_object(self):
        self_dict = vars(self)
        content = self_dict['richContent'].copy()[0]
        for ci in range(len(content)):
            if hasattr(content[ci], 'to_object'):
                content[ci] = content[ci].to_object()
            
        return self_dict

    def add_content(self, *ctnt):
        self.richContent[0].extend(ctnt)

    def __call__(self):
        return self.to_object()


class ContentType:
    
    def __init__(self, type_: str):
        self.type = type_
        
    def to_object(self):
        self_dict = vars(self)
        for key, value in self_dict.items():
            if isinstance(value, list):
                value_list = value.copy()
                for li in range(len(value_list)):
                    if hasattr(value_list[li], 'to_object'):
                        value_list[li] = value_list[li].to_object()
                self_dict[key] = value_list
            if hasattr(value, 'to_object'):
                self_dict[key] = value.to_object()
        return self_dict
    
    def __call__(self):
        return self.to_object()

    
class Info(ContentType):
    
    def __init__(self, title: str, subtitle: str, image_url, anchor_url):
        super().__init__('info')
        self.title = title
        self.subtitle = subtitle
        self.image = {
            'rawUrl': image_url
        } 
        self.anchor = {
            'href': anchor_url
        }
        if not image_url:
            del self.image
        if not anchor_url:
            del self.anchor


class Description(ContentType):
    
    def __init__(self, title: str, *texts: str):
        super().__init__('description')
        self.title = title
        # 2024-01-31: updated [texts] to list(texts) which converts the tuple 
        # into a list.
        self.text = list(texts)

        
class Image(ContentType):
    
    def __init__(self, image_url: str, accessibility_text: str):
        super().__init__('image')
        self.rawUrl = image_url
        self.accessibilityText = accessibility_text
        

class Video(ContentType):
    # Omitted due to complexity
    ...
    

class Button(ContentType):
    
    def __init__(self, icon_type: str, icon_color: str, mode: str, text: str, anchor_url: str, event: str = ""):
        super().__init__('button')
        self.icon = {
            'type': icon_type,
            'color': icon_color
        }
        self.mode = mode
        self.text = text
        self.anchor = {
            'href': anchor_url
        }
        self.event = {
            'event': event
        }
        
class Chip:
    
    def __init__(self, mode: str, text: str, image_url: str, anchor_url: str):
        self.mode = mode
        self.text = text
        self.image = {
            'rawUrl': image_url
        }
        self.anchor = {
            'href': anchor_url
        }
        
    def to_object(self):
        return vars(self)
        
    
class Chips(ContentType):
    
    def __init__(self, *chips: Chip):
        super().__init__('chips')
        self.options = list(chips)
        

class Citation:
    
    def __init__(self, title: str, subtitle: str, anchor_url: str, anchor_target: str = None):
        self.title = title
        self.subtitle = subtitle
        self.anchor = {
            'href': anchor_url
        }
        if anchor_target :
            self.anchor['target'] = anchor_target
            
    def to_object(self):
        return vars(self)


class Citations(ContentType):
    
    def __init__(self, *chips: Chip):
        super().__init__('match_citations')
        self.citations = list(chips)



import argparse
import urllib.request

FILES = {
    'Dockerfile': 'https://gitlab.com/cloud-colosseum/ez/-/raw/main/cx/Dockerfile',
    'cloudbuild.yaml': 'https://gitlab.com/cloud-colosseum/ez/-/raw/main/cx/cloudbuild.yaml',
    'main.py': 'https://gitlab.com/cloud-colosseum/ez/-/raw/main/cx/main.py'
}

class EZ_CLI:

    def __init__(self):
        self.command_tree = {
            'generate': {
                'scaffolding': self.__scaffolding,
                'quickstart': self.__quickstart
            }
        }
        self.parser = self.setup_parser()

    def setup_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers()
        generate = subparsers.add_parser('generate')
        generate_subparsers = generate.add_subparsers(dest='generate')
        for subcmd in self.command_tree['generate']:
            generate_subparsers.add_parser(subcmd)
        return parser

    def __scaffolding(self):
        files = {'Dockerfile', 'cloudbuild.yaml'}
        for file, url in FILES.items():
            if file not in files:
                continue
            print(f'[ez-cli]: downloading {file}')
            with open(file, 'wb') as dest, urllib.request.urlopen(url) as response:
                dest.write(response.read())
        print()

    def __quickstart(self):
        files = {'Dockerfile', 'cloudbuild.yaml', 'main.py'}
        for file, url in FILES.items():
            if file not in files:
                continue
            print(f'[ez-cli]: downloading {file}')
            with open(file, 'wb') as dest, urllib.request.urlopen(url) as response:
                dest.write(response.read())

    def __call__(self):
        args = self.parser.parse_args()
        subp = self.command_tree['generate'].get(args.generate)
        subp()

cli = EZ_CLI()






from google.cloud import dialogflowcx as cx

from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

# 2023-12-11 
# - Now with docstrings!

class WebhookResponse:

    """
    Class to build a WebhookResponse object.

    Args:
        None.
    """

    def __init__(self):
        self.__message: cx.WebhookResponse = cx.WebhookResponse(
            fulfillment_response=self.FulfillmentResponse(messages=[])
        )

    # gRPC Message attributes exposed as properties;
    # Easier on traditional linters and formatters.

    @property
    def fulfillment_response(self) -> cx.WebhookResponse.FulfillmentResponse:
        """
        Returns the fulfillment response.

        Args:
            None.

        Returns:
            cx.WebhookResponse.FulfillmentResponse: The fulfillment response.
        """
        return self.__message.fulfillment_response

    @property
    def page_info(self) -> cx.PageInfo:
        """
        Returns the page info.

        Args:
            None.

        Returns:
            cx.PageInfo: The page info.
        """
        return self.__message.page_info

    @property
    def session_info(self) -> cx.SessionInfo:
        """
        Returns the session info.

        Args:
            None.

        Returns:
            cx.SessionInfo: The session info.
        """
        return self.__message.session_info

    @property
    def payload(self) -> Struct:
        """
        Returns the payload.

        Args:
            None.

        Returns:
            Struct: The payload.
        """
        return self.__message.payload

    @property
    def FulfillmentResponse(self) -> cx.WebhookResponse.FulfillmentResponse:
        """
        Returns the FulfillmentResponse class.

        Args:
            None.

        Returns:
            cx.WebhookResponse.FulfillmentResponse: The FulfillmentResponse class.
        """
        return cx.WebhookResponse.FulfillmentResponse

    @property
    def ResponseMessage(self) -> cx.ResponseMessage:
        """
        Returns the ResponseMessage class.

        Args:
            None.

        Returns:
            cx.ResponseMessage: The ResponseMessage class.
        """
        return cx.ResponseMessage

    @property
    def Text(self) -> cx.ResponseMessage.Text:
        """
        Returns the Text class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.Text: The Text class.
        """
        return cx.ResponseMessage.Text

    @property
    def ConversationSuccess(self) -> cx.ResponseMessage.ConversationSuccess:
        """
        Returns the ConversationSuccess class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.ConversationSuccess: The ConversationSuccess class.
        """
        return cx.ResponseMessage.ConversationSuccess

    @property
    def OutputAudioText(self) -> cx.ResponseMessage.OutputAudioText:
        """
        Returns the OutputAudioText class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.OutputAudioText: The OutputAudioText class.
        """
        return cx.ResponseMessage.OutputAudioText

    @property
    def LiveAgentHandoff(self) -> cx.ResponseMessage.LiveAgentHandoff:
        """
        Returns the LiveAgentHandoff class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.LiveAgentHandoff: The LiveAgentHandoff class.
        """
        return cx.ResponseMessage.LiveAgentHandoff

    @property
    def PlayAudio(self) -> cx.ResponseMessage.PlayAudio:
        """
        Returns the PlayAudio class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.PlayAudio: The PlayAudio class.
        """
        return cx.ResponseMessage.PlayAudio

    @property
    def TelephonyTransferCall(self) -> cx.ResponseMessage.TelephonyTransferCall:
        """
        Returns the TelephonyTransferCall class.

        Args:
            None.

        Returns:
            cx.ResponseMessage.TelephonyTransferCall: The TelephonyTransferCall class.
        """
        return cx.ResponseMessage.TelephonyTransferCall

    def add_response(self, response_message: cx.ResponseMessage):
        """Adds a response message to the WebhookResponse object.

        Args:
            response_message: The response message to add.

        Returns:
            The WebhookResponse object.
        """
        self.fulfillment_response.messages.append(response_message)
        return self

    def add_text_response(self, *texts, channel=""):
        """Adds a text response message to the WebhookResponse object.

        Args:
            *texts: The text messages to add.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        text = self.Text(text=texts)
        response_message = self.ResponseMessage(text=text, channel=channel)
        self.add_response(response_message)
        return self

    def add_conversation_success(self, metadata: dict, channel=""):
        """Adds a conversation success response message to the WebhookResponse object.

        Args:
            metadata: The metadata to include in the response message.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        conversation_success = self.ConversationSuccess(metadata=metadata)
        response_message = self.ResponseMessage(conversation_success=conversation_success, channel=channel)
        self.add_response(response_message)
        return self

    def add_payload_response(self, payload: dict, channel=""):
        """Adds a payload response message to the WebhookResponse object.

        Args:
            payload: The payload to include in the response message.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        # ResponseMessage instantiation with value of Payload handles this automatically
        # This is the "mapping" interface; no need for Struct and ParseDict
        response_message = self.ResponseMessage(payload=payload, channel=channel)
        self.add_response(response_message)
        return self

    def add_ssml_response(self, ssml: str, channel=""):
        """Adds a SSML response message to the WebhookResponse object.

        Args:
            ssml: The SSML to include in the response message.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        output_audio_text = self.OutputAudioText(ssml=ssml)
        response_message = cx.ResponseMessage(output_audio_text=output_audio_text, channel=channel)
        self.add_response(response_message)
        return self

    def add_play_audio(self, audio_uri: str, channel=""):
        """Adds a play audio response message to the WebhookResponse object.

        Args:
            audio_uri: The audio URI to include in the response message.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        play_audio = self.PlayAudio(audio_uri=audio_uri)
        response_message = cx.ResponseMessage(play_audio=play_audio, channel=channel)
        self.add_response(response_message)
        return self

    def add_live_agent_handoff(self, metadata: dict, channel=""):
        """Adds a live agent handoff response message to the WebhookResponse object.

        Args:
            metadata: The metadata to include in the response message.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        live_agent_handoff = self.LiveAgentHandoff(metadata=metadata)
        response_message = cx.ResponseMessage(live_agent_handoff=live_agent_handoff, channel=channel)
        self.add_response(response_message)
        return self

    def add_telephony_transfer_call(self, phone_number: str, channel=""):
        """Adds a telephony transfer call response message to the WebhookResponse object.

        Args:
            phone_number: The phone number to transfer the call to.
            channel: The channel to send the response message to.

        Returns:
            The WebhookResponse object.
        """
        telephony_transfer_call = self.TelephonyTransferCall(phone_number=phone_number)
        response_message = cx.ResponseMessage(telephony_transfer_call=telephony_transfer_call, channel=channel)
        self.add_response(response_message)
        return self

    def add_session_parameters(self, parameters: dict):
        """Adds session parameters to the WebhookResponse object.

        Args:
            parameters: The session parameters to add.

        Returns:
            The WebhookResponse object.
        """
        session_info = cx.SessionInfo(parameters=parameters)
        self.__message.session_info = session_info
        return self

    def set_payload(self, payload: dict):
        """Sets the payload of the WebhookResponse object.

        Args:
            payload: The payload to set.

        Returns:
            The WebhookResponse object.
        """
        self.__message.payload = payload
        return self

    def set_transition(self, target: str):
        """Sets the transition of the WebhookResponse object.

        Args:
            target: The target of the transition.

        Returns:
            The WebhookResponse object.
        """
        target_length = len(target.split('/'))
        if target_length == 10:
            self.__message.target_page = target
        elif target_length == 8:
            self.__message.target_flow = target
        return self

    # JSON Encoding methods.  These are primarily for testing and logging.
    def to_dict(self) -> dict:
        """Converts the WebhookResponse object to a dictionary.

        Args:
            None

        Returns:
            A dictionary representation of the WebhookResponse object.
        """
        return MessageToDict(self.__message._pb, including_default_value_fields=True)

    @property
    def as_dict(self) -> dict:
        """Returns the WebhookResponse object as a dictionary.

        Args:
            None

        Returns:
            A dictionary representation of the WebhookResponse object.
        """
        return self.to_dict()



import json 

from google.cloud import dialogflowcx as cx
from google.protobuf.json_format import MessageToDict

from httpx import Client

def wh_request(tag: str, 
    parameters: dict = None, 
    text: str = None, 
    transcript: str = None
):
    wh_request = cx.WebhookRequest({}, ignore_unknown_fields=True)
    wh_request.fulfillment_info.tag = tag
    if parameters:
        wh_request.session_info.parameters = parameters
    if text:
        wh_request.text = text
    if transcript:
        wh_request.transcript = transcript
    
    return MessageToDict(wh_request._pb, including_default_value_fields=True)


with Client() as client:
    server_url = 'http://localhost:9080/'
    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('hello_world'))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))
    
    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('user_query', text='hi there!'))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))

    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('user_query', transcript='hi there!'))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))

    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('session_parameters', parameters={'message': 'ezcx makes cx ez'}))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))

    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('update_session_parameters', parameters={'message': 'ezcx makes cx ez'}))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))

    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('telephony_transfer'))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))

    print('\t--- --- ---\t')
    response = client.post(server_url, json=wh_request('multichannel_responses'))
    if response.status_code == 200:
        print(json.dumps(response.json(), indent=2))
    


from google.cloud import dialogflowcx as cx
from google.protobuf.json_format import ParseDict
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct

# 2023-12-11 
# - Now with docstrings!

# 2023-12-01
# - Dialogflow CX WebhookRequest object has changed with breaking changes
# - addition of "query" helper method to work through incoming payloads created.

# 2023-10-28
# - Replacement of self.request with self.__message to represent the proto message
# - Addition of __getattr__ which fetches from the self.__message now

class WebhookRequest:

    """A facade representing the Dialogflow CX WebhookRequest object.

    This class provides a convenient way to access the fields of a
    WebhookRequest object.

    Args:
        body: A dictionary representing the JSON payload of the webhook request.
    """

    def __init__(self, body: dict = {}):
        self.__message = cx.WebhookRequest()
        # When a webhook comes in as JSON, we use ParseDict to convert JSON to Protobuf
        ParseDict(body, self.__message._pb, ignore_unknown_fields=True)
        self.__query = None
        self.__origin = None

    # gRPC Message attributes exposed as properties;
    # Easier on traditional linters and formatters.

    @property
    def detect_intent_response_id(self) -> str:
        """The unique identifier of the DetectIntentResponse that this webhook
        request is derived from.

        Returns:
            A string representing the unique identifier of the DetectIntentResponse.
        """
        return self.__message.detect_intent_response_id

    @property
    def language_code(self) -> str:
        """The language code of the query that the agent received.

        Returns:
            A string representing the language code of the query.
        """
        return self.__message.language_code

    @property
    def fulfillment_info(self) -> cx.WebhookRequest.FulfillmentInfo:
        """Information about the fulfillment that triggered this webhook call.

        Returns:
            A FulfillmentInfo object representing the fulfillment information.
        """
        return self.__message.fulfillment_info

    @property
    def intent_info(self) -> cx.WebhookRequest.IntentInfo:
        """Information about the intent matched by the agent.

        Returns:
            An IntentInfo object representing the intent information.
        """
        return self.__message.intent_info

    @property
    def session_info(self) -> cx.SessionInfo:
        """Information about the session for which this webhook request is made.

        Returns:
            A SessionInfo object representing the session information.
        """
        return self.__message.session_info

    @property
    def messages(self) -> list:
        """The list of messages that the agent has sent to the user.

        Returns:
            A list of Message objects representing the messages.
        """
        return self.__message.messages

    @property
    def payload(self) -> Struct:
        """The payload of the webhook request.

        Returns:
            A Struct object representing the payload.
        """
        return self.__message.payload

    @property
    def sentiment_analysis_result(self) -> cx.SentimentAnalysisResult:
        """The sentiment analysis result of the query.

        Returns:
            A SentimentAnalysisResult object representing the sentiment analysis result.
        """
        return self.__message.sentiment_analysis_result

    # Helper properties.  These class properties make basic access easier
    @property
    def tag(self) -> str:
        """The tag of the fulfillment that triggered this webhook call.

        Returns:
            A string representing the tag of the fulfillment.
        """
        return self.__message.fulfillment_info.tag

    @property
    def session(self) -> str:
        """The name of the session for which this webhook request is made.

        Returns:
            A string representing the name of the session.
        """
        return self.__message.session_info.session

    @property
    def session_id(self) -> str:
        """The unique identifier of the session for which this webhook request is made.

        Returns:
            A string representing the unique identifier of the session.
        """
        return self.session.split('/')[-1]

    @property
    def session_parameters(self):
        """The session parameters for this webhook request.

        Returns:
            A dictionary representing the parameters of the session.
        """
        return MessageToDict(
            self.__message.session_info._pb,
            including_default_value_fields=True,
        ).get('parameters')

    @property
    def query(self):
        """The query that the user sent to the agent.

        Returns:
            A tuple containing the query and the origin of the query.
        """
        if self.__query and self.__origin:
            return self.__query, self.__origin

        r = self.__message

        # q is for query, o is for origin
        q, o = '', ''
        if r.text:
            q = r.text
            o = 'text'
        elif r.trigger_event:
            q = r.trigger_intent
            o = 'trigger_intent'
        elif r.transcript:
            q = r.transcript
            o = 'transcript'
        elif r.trigger_event:
            q = r.trigger_event
            o = 'trigger_event'
        elif r.dtmf_digits:
            q = r.dtmf_digits
            o = 'dtmf_digits'
        else:
            ...

        self.__query = q
        self.__origin = o
        return self.__query, self.__origin

    @property
    def origin(self) -> str:
        """The origin of the query that the user sent to the agent.

        Returns:
            A string representing the origin of the query.
        """
        if not self.__origin:
            self.query
        return self.__origin

    # JSON Encoding methods.  These are primarily for testing and logging.
    def to_dict(self) -> dict:
        """Converts the WebhookRequest object to a dictionary.

        Returns:
            A dictionary representing the WebhookRequest object.
        """
        return MessageToDict(self.__message._pb, including_default_value_fields=True)

    @property
    def as_dict(self) -> dict:
        """Returns the WebhookRequest object as a dictionary.

        Returns:
            A dictionary representing the WebhookRequest object.
        """
        return self.to_dict()